import yaml
import importlib

import collections
import functools
from schema import Schema, Or, SchemaError
from .context import Context
from manifest_edit import libfmp4

FORMAT_MPD  = "mpd"
FORMAT_M3U8_MAIN = "m3u8_main"
FORMAT_M3U8_MEDIA = "m3u8_media"

_supported_formats = [FORMAT_MPD, FORMAT_M3U8_MAIN, FORMAT_M3U8_MEDIA]


class Config:

    _schema = Schema(
        {
            Or(*_supported_formats):
            [
                {
                    # plugins path: configuration or empty in case you
                    # want to receive it from a previous plugin
                    str: Or(list, dict, None)
                }
            ]
        }
    )

    @ staticmethod
    def validate(config):
        return Config._schema.validate(config)

# From https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize


class Memoized:
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)


class Pipeline:

    # TODO needs to be shared and thus elsewhere. Notice that support for
    # m3u8 is missing at the moment
    SUPPORTED_MANIFEST_FORMATS = _supported_formats

    @ staticmethod
    @ Memoized
    def build_pipeline(yaml_file):
        '''Static method to build a pipeline from a yaml configuration file.
        '''
        with open(yaml_file, 'r') as stream:
            try:
                # Trick: using stream.read() to improve error reporting
                #  by pyyaml! Buffers are better handled than streams.
                pipeline_def_raw = yaml.safe_load(stream.read())
            except yaml.YAMLError as e:
                Context.log_error('Invalid yaml configuration in {}:'
                                  '  {}\n'
                                  .format(yaml_file, e)
                                  )
                raise e

            try:
                # will raise exception if overall yml file config is invalid
                pipeline_def = Config.validate(pipeline_def_raw)
            except SchemaError as e:
                Context.log_error(
                    'Invalid Yaml configuration:\n    {}'.format(e.autos[-1]))
                raise e

            # A config file can contain a pipeline definition per manifest
            # format

            manifest_format, plugin_def = next(
                iter(pipeline_def.items())
            )
            plugins_list = []

            for p in plugin_def:
                plugin_name, plugin_config = next(iter(p.items()))
                # Here you need to create the plugin
                try:
                    # Remember that python imports just once, no matter
                    # how many times you call this
                    plugin = importlib.import_module(plugin_name)
                    importlib.invalidate_caches()
                except ImportError as e:
                    Context.log_error(
                        'Unable to find python plugin defined in {}\n'
                        ' Exception {}'.
                        format(yaml_file, e)
                    )
                    raise e
                plugin_class_instance = getattr(plugin, "Plugin")()
                # Base class plugin method to specialize processing
                #  function on the specific format
                plugin_class_instance.onCreate(
                    manifest_format, plugin_config)
                try:
                    if plugin_config is None:
                        # This is a plugin that will receive its config
                        #  from another plugin and not from the yml file.
                        plugins_list.append(plugin_class_instance)
                    elif plugin_class_instance.validate_config():
                        # This is one that gets its config from a yml file
                        # so we had to know if that is correct
                        plugins_list.append(plugin_class_instance)
                except SchemaError as e:
                    errors = set([err.rsplit(
                        'validate')[-1] for err in e.autos if err is not None and 'validate' in err])
                    error_messages = ''
                    for err in errors:
                        error_messages += err+'\n'
                    Context.log_error(
                        'Invalid configuration for plugin {}:\n    {}'.format(plugin.__name__, error_messages))
                    raise e

        return Pipeline(
            manifest_format, plugins_list,
            yaml.dump(pipeline_def, default_flow_style=False)
        )

    def __init__(self, manifest_format, plugins_list, yaml_configuration):
        self.format = manifest_format
        self._pipeline = plugins_list
        self.yaml_configuration = yaml_configuration

    def run(self, input_manifest=None, **kwargs):
        # Process across the pipeline
        # The storage dictionary can be used by plugins to propagate anything
        #  down the pipeline to subsequent plugins
        storage = {}
        for plugin in self._pipeline:
            # Expecially after having memoized the pipeline build, it is
            #  necessary to explicitly call constructors to make sure that
            #  each subsequent invocation of the pipeline finds plugins in the
            #  expected initialized state.
            plugin.__init__()
            plugin.process(manifest=input_manifest, storage=storage)
