import abc
from manifest_edit.pipeline import (
    FORMAT_MPD,
    FORMAT_M3U8_MAIN,
    FORMAT_M3U8_MEDIA
)


class Mpd():
    supported_format = FORMAT_MPD


class M3u8Main():
    supported_format = FORMAT_M3U8_MAIN


class M3u8Media():
    supported_format = FORMAT_M3U8_MEDIA


class BasePlugin(abc.ABC):
    '''
    BasePlugin is the simplest base class that a Plugin can inherit from.
    It's the class you should use if your plugin is very simple or if you
    want total freedom on the design of your Plugin subclass.

    If you need to write a plugin that acts on dynamically selected
    representations, adaptationSets or periods, you should have a look
    at ManifestIteratorPlugin base class instead.

    '''

    def onCreate(self, manifest_format, config):

        if not hasattr(self, 'supported_format'):
            raise Exception('Error in plugin creation! The plugin does not'
                            ' specify a sypported format. Please inherit'
                            ' from one of the supported format classes'
                            ' from the manifest_edit.plugin subpackage.')

        elif self.supported_format != manifest_format:
            raise Exception('Error in plugin creation!'
                            ' You have included a manifest with format {}'
                            ' in a {} pipeline!'.format(
                                self.supported_format,
                                manifest_format))

        self._config = config

    # Plugins MUST reimplement at least these methods to be valid
    # This is invoked at creation time and is necessary to validate
    #  that the .yaml configuration provided for the specific plugin
    #  is valid. The recommended way to implement this is using Schema.
    # Doing it here has the advantages of having the actual manifest processing
    #  code (i.e mpd()) cleaner (you can give for granted that your config
    #  dict will have the expected keys,values). Plus, configuration validation
    #  can be computationally expensive so it is just right to do it once
    #  when the .yaml config file is loaded.

    def config(self, storage):
        # The "config" method of a BasePlugin will return a dictionary type
        if self._config is not None:
            return self._config
        else:
            # Plugins will pass a configuration down the pipeling by putting it
            #  into a {<plugin name> : {<config>}} dictionary. The split is
            #  there to remove the path. They would use the name of the plugin
            #  that must receive the configuration.
            return storage.get(self._name.split('.')[-1], {})

    def validate_config(self):
        self._config = self.schema().validate(self._config)
        return self._config

    @property
    @abc.abstractmethod
    def schema(self):
        raise NotImplementedError

    @abc.abstractmethod
    def process(self, manifest, storage):
        raise NotImplementedError
