from manifest_edit.plugin import ManifestIteratorPlugin, Mpd
from manifest_edit.context import Context
from schema import Schema, Or
import re


class Plugin(Mpd, ManifestIteratorPlugin):
    '''
    supplemental_property_remove plugin

    This plugin will remove one or more SupplementalProperty elements from a
    specified position in a manifest.

    It can be used to remove all SupplementalProperty elements, or only those with
    a specific scheme or value. Schemes and values are checked using regular
    expressions.

    For example, to remove all from any Adaptation Set:

    .. code-block:: yaml

        periods:
        - '*': '.*'
          adaptationSets:
          - '*': '.*'
            supplementalPropertyScheme: '.*'
            supplementalPropertyValue: '.*'

    Or to only remove a specific property:

    .. code-block:: yaml

        periods:
        - '*': '.*'
          adaptationSets:
          - '*': '.*'
            supplementalPropertyScheme: 'urn:dvb:dash:lowlatency:critical:2019'
            supplementalPropertyValue: 'true'

    '''

    _name = __name__

    _keys = ['supplementalPropertyScheme', 'supplementalPropertyValue']

    def schema(self):
        return Schema({
            self._keys[0]: str,
            self._keys[1]: Or(str, int)
        })

    def removeSupplementalProperties(self, manifest, storage):
        for ess_prop_config, element in self.config(manifest, storage):
            properties_to_remove = []

            for prop in element.supplementalProperties:
                if (re.fullmatch(ess_prop_config['supplementalPropertyScheme'],
                                 prop.schemeIdUri)
                    and
                    re.fullmatch(ess_prop_config['supplementalPropertyValue'],
                                 prop.value)):
                    properties_to_remove.append(prop)

            for prop in properties_to_remove:
                element.supplementalProperties.remove(prop)

    def process(self, manifest, storage):
        self.removeSupplementalProperties(manifest, storage)